import datetime
import pytest
import sys
import unittest
from django.core.management import call_command
from io import StringIO

from ..models import BlockedIP


@pytest.mark.django_db
class CommandsTest(unittest.TestCase):
    def setUp(self):
        BlockedIP.objects.all().delete()

    def tearDown(self):
        BlockedIP.objects.all().delete()

    def test_clean(self):
        two_days_ago = datetime.datetime.now() - datetime.timedelta(days=2)
        BlockedIP.objects.create(ip="1.1.1.1", cooldown=1, last_seen=two_days_ago)
        BlockedIP.objects.create(ip="2.2.2.2", cooldown=3, last_seen=two_days_ago)
        # We now have two IPs
        assert BlockedIP.objects.count() == 2
        # Running clean_blocklist with --dry-run doesn't delete one
        call_command("clean_blocklist", verbosity=0, dry_run=1)
        self.assertEqual(BlockedIP.objects.count(), 2)
        # Running clean_blocklist does delete one
        call_command("clean_blocklist", verbosity=0)
        self.assertEqual(BlockedIP.objects.count(), 1)
        # The one that should still be there, is still there
        self.assertTrue(BlockedIP.objects.filter(ip="2.2.2.2").exists())

    def test_add(self):
        call_command("update_blocklist", "3.3.3.3", verbosity=0)
        self.assertTrue(BlockedIP.objects.filter(ip="3.3.3.3").exists())

    def test_remove(self):
        BlockedIP.objects.create(ip="4.4.4.4")
        call_command("remove_from_blocklist", "4.4.4.4", verbosity=0)
        self.assertEqual(BlockedIP.objects.filter(ip="4.4.4.4").count(), 0)

    def test_add_invalid(self):
        out = StringIO()
        sys.stdout = out
        bad_ip = "foo"
        call_command("update_blocklist", bad_ip, verbosity=0)
        self.assertIn("Invalid", out.getvalue())

    def test_update(self):
        BlockedIP.objects.create(ip="5.5.5.5", reason="R1", cooldown=1)
        call_command("update_blocklist", "5.5.5.5", reason="R2", cooldown=2)
        entry = BlockedIP.objects.get(ip="5.5.5.5")
        self.assertEqual(entry.reason, "R2")
        self.assertEqual(entry.cooldown, 2)

    def test_report(self):
        out = StringIO()
        sys.stdout = out
        today = datetime.datetime.now()
        yesterday = today - datetime.timedelta(days=1)
        BlockedIP.objects.create(ip="7.7.7.7", first_seen=yesterday, last_seen=today)
        BlockedIP.objects.create(ip="8.8.8.8", first_seen=yesterday, last_seen=today, tally=240)
        call_command("report_blocklist")
        result = out.getvalue()
        self.assertIn("7.7.7.7 -- 0 blocks", result)
        self.assertIn("10 per hour", result)

    def test_reason_report(self):
        out = StringIO()
        sys.stdout = out
        reasons = ["A", "B"]
        for n, reason in enumerate(reasons):
            BlockedIP.objects.create(ip=f"{n}.{n}.{n}.{n}", reason=reason)
        call_command("report_blocklist", reason=reasons[0])
        result = out.getvalue()
        assert BlockedIP.objects.count() == 2
        self.assertIn("Entries in blocklist: 1", result)



"""Remove IPs from the blocklist if they have been inactive for the required cooldown."""
import datetime
import logging

from django.db.models import Min
from django.core.management.base import BaseCommand

from ...models import BlockedIP

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--dry-run", action="store_true", help="Preview the removal rather than performing it."
        )

    help = __doc__

    def handle(self, *args, **options):
        dry_run = options.get("dry_run")
        self.verbosity = options.get("verbosity")
        total_at_start = BlockedIP.objects.count()
        deletion_count = 0
        shortest_cooldown_in_db = BlockedIP.objects.aggregate(shortest=Min("cooldown"))["shortest"]
        latest_possible_timestamp_of_expired_entries = datetime.datetime.now() - datetime.timedelta(
            days=shortest_cooldown_in_db
        )
        for entry in BlockedIP.objects.filter(last_seen__lte=latest_possible_timestamp_of_expired_entries):
            if entry.has_expired():
                deletion_count += 1
                if not dry_run:
                    entry.delete()
        if dry_run:
            message = f"Would have removed {deletion_count} IPs."
        else:
            message = (
                f"Removed {deletion_count} IPs from blocklist; {total_at_start - deletion_count} remain."
            )
        logger.info(message)



"""Update blocklist with new IPs, or set metadata on existing IPs."""
import logging

from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand
from django.core.validators import validate_ipv46_address

from ...models import BlockedIP
from ...utils import COOLDOWN


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("ips", nargs="+", type=str, help="IPs (space-separated)")
        parser.add_argument(
            "--cooldown",
            help=f"Days with no requests before IP is dropped from blocklist (default: {COOLDOWN})",
        )
        parser.add_argument("--reason", help="'reason' field value for these IPs", default="")

    help = __doc__

    def handle(self, *args, **options):
        ips = options.get("ips")
        cooldown = options.get("cooldown")
        reason = options.get("reason")
        for ip in ips:
            try:
                validate_ipv46_address(ip)
            except ValidationError:
                print(f"Invalid IP: {ip}")
                continue
            entry, created = BlockedIP.objects.get_or_create(ip=ip)
            if reason:
                entry.reason = reason
            if cooldown:
                entry.cooldown = cooldown
            entry.save()
            print(f"{'Created' if created else 'Updated'} entry for {ip}")



from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="BlockedIP",
            fields=[
                ("ip", models.GenericIPAddressField(primary_key=True, serialize=False, verbose_name="IP")),
                ("first_seen", models.DateTimeField(db_index=True, default=django.utils.timezone.now)),
                ("last_seen", models.DateTimeField(blank=True, db_index=True, null=True)),
                ("reason", models.CharField(blank=True, db_index=True, default="", max_length=255)),
                (
                    "tally",
                    models.IntegerField(
                        default=1, help_text="Number of times this IP has been blocked since first_seen"
                    ),
                ),
                (
                    "cooldown",
                    models.IntegerField(
                        default=7,
                        help_text="Cooldown period; number of days with no connections before IP is dropped from blocklist",
                    ),
                ),
            ],
            options={
                "verbose_name": "blocked IP",
                "ordering": ["-last_seen", "ip"],
                "get_latest_by": "first_seen",
            },
        ),
    ]



"""Generate fake blocklist entries for testing. Only generates non-routable IPs."""
import datetime
import sys
from random import randint

from django.core.management.base import BaseCommand

from ...models import BlockedIP


def random_fake_ip() -> str:
    template = "192.168.{}.{}"
    return template.format(randint(0, 255), randint(0, 255))


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument("--quantity", type=int, help="Number of fake IPs to generate")

    def handle(self, *args, **options):
        if (quantity := options.get("quantity")) > (max := 256**2):
            sys.exit(f"Can't generate more than {max}, sorry.")
        while quantity:
            entry, created = BlockedIP.objects.get_or_create(ip=random_fake_ip())
            if created:
                entry.tally = randint(1, 100)
                entry.reason = f"Fake reason {randint(1,3)}"
                entry.first_seen = datetime.datetime.now() - datetime.timedelta(randint(1, 10))
                entry.last_seen = entry.first_seen + datetime.timedelta(days=randint(1, entry.cooldown - 1))
                entry.save()
                quantity -= 1
        print(f"Generated {options.get('quantity')} BlockedIP entries.")



from django.core.cache import cache
from django.db.utils import IntegrityError
from django.http import HttpRequest
from django.test import TestCase

from ..utils import (
    CACHE_KEY,
    add_to_blocklist,
    get_blocklist,
    remove_from_blocklist,
    user_ip_from_request,
)


class UtilsTests(TestCase):
    def setUp(self):
        cache.delete(CACHE_KEY)

    def tearDown(self):
        cache.delete(CACHE_KEY)

    def test_get_list(self):
        cache.set(CACHE_KEY, ["1.1.1.1"])
        result = get_blocklist()
        self.assertEqual(result, {"1.1.1.1"})

    def test_add(self):
        ips = {"2.2.2.2", "3.3.3.3"}
        add_to_blocklist(ips, reason="whatever")
        result = get_blocklist()
        self.assertEqual(result, ips)

    def test_add_duplicate(self):
        ip = "9.9.9.9"
        add_to_blocklist(ip)
        try:
            add_to_blocklist(ip)
        except IntegrityError:
            self.fail("add_to_blocklist failed to handle duplicate IP")

    def test_remove(self):
        ip_as_list = ["1.1.1.1"]
        cache.set(CACHE_KEY, ip_as_list)
        assert get_blocklist() == set(ip_as_list)
        remove_from_blocklist(ip_as_list)
        self.assertEqual(get_blocklist(force_cache=True), set())

    def test_get_user_ip(self):
        request = HttpRequest()
        result = user_ip_from_request(request)
        self.assertEqual(result, "")
        ip = "4.4.4.4"
        for key in ["HTTP_X_REAL_IP", "REMOTE_ADDR"]:
            request.META[key] = ip
            self.assertEqual(user_ip_from_request(request), ip)
            del request.META[key]
        request.META["HTTP_X_FORWARDED_FOR"] = "5.5.5.5,6.6.6.6"
        result = user_ip_from_request(request)
        self.assertEqual(result, "5.5.5.5")



from django.contrib import admin
from django.http import HttpResponseRedirect

from .models import BlockedIP


@admin.display(description="Reason")
def reason_truncated(entry):
    return entry.reason[:20] + ("..." if len(entry.reason) > 20 else "")


@admin.display(description="Cooldown")
def cooldown(entry):
    return f"{entry.cooldown} days"


@admin.action(permissions=["view"])
def look_up_first_selected_IP(modeladmin, request, queryset):
    obj = queryset[0]
    return HttpResponseRedirect(f"https://whatismyipaddress.com/ip/{obj.ip}")


class BlockedIPAdmin(admin.ModelAdmin):
    list_display = ["ip", "first_seen", "last_seen", "tally", cooldown, reason_truncated]
    list_filter = ["first_seen", "last_seen", "cooldown", "reason"]
    search_fields = ["ip", "reason"]
    actions = [look_up_first_selected_IP]

    class Meta:
        model = BlockedIP


admin.site.register(BlockedIP, BlockedIPAdmin)
